import numpy as np
from evspy.single_stage_model.creep_swelling_model import power_law

def swelling_law(erates,OCR,OCRrate,Calphahats_real,dt,CalphaNC,m1,b1,m2,b2,sigmarate,erate,reset):    
    '''The swelling law defines the evolution of the swelling strain rate. A minimum strain rate is imposed as an activation condition at the start of a load step.
    

    Parameters
    ----------
    erates : float
        swelling strain rate
    OCR : float
        Overconsolidation ratio
    OCRrate : float
        Rate of change of OCR
    Calphahats_real : float
        Current swelling Calpha
    dt : float
        Time step
    CalphaNC : float
        Calpha at NC
    m1 : float
        Slope of the power law for Calphahats
    b1 : float
        Intercept of the power law Calphahats
    m2 : float
        Slope of the power law the strain rate
    b2 : float
        Intercept of the power law the strain rate
    sigmarate : float
        Rate of the stress
    erate : float
        Void ratio strain rate
    reset : int
        list to control a reset - minimum strain rate is only enforced at start of load step

    Returns
    -------
    tuple
        [description]
    '''
    OCRmin=1.05
    OCR=np.clip(OCR,OCRmin,np.infty)
    OCRrate=OCRrate*(OCR>OCRmin)
    
    try:
        # If the model is coupled; all variables are defined as lists instead of floats and conditions have to be looped. 
        min_erates=np.zeros((len(reset)))
        for x in range(len(reset)):
            if (reset[x]<2) and (OCR[x]>OCRmin).all() and ((sigmarate[x]<0) and (erate[x]>0)):
                min_erates[x] = power_law(OCRmin,b=b2,m=m2)#*((sigmarate[x]<0)&(erate[x]>0))
                if x == 2: # Only valid at soil element 2
                    Calphahats_real = np.clip(Calphahats_real,
                                                # Clip at Calphahat_s at OCR = 3 if load is changing and soil is straining to avoid high sensitivity during unloading
                                                CalphaNC*power_law(3,b=b1,m=m1)*((np.average(sigmarate)<-1e-20)&(np.average(erate)>1e-20)),np.infty)
                reset[x]+=1
                
            else:
                min_erates[x] = erates[x]/2
    except:
        # If the model is decoupled, all variables are signle floats. 
        if (reset<2) and (OCR>OCRmin).all():
            min_erates = power_law(OCRmin,b=b2,m=m2)*((np.average(sigmarate)<0)&(np.average(erate)>0))
            reset+=1
        else:
            min_erates = erates/2
    
    Calphahats_real = np.clip(Calphahats_real,CalphaNC*power_law(3,b=b1,m=m1)*((np.average(sigmarate)<-1e-20)&(np.average(erate)>1e-20)),np.infty)
    return np.clip(erates+m2*OCRrate/(OCR-1)*erates*dt-erates*erates/(Calphahats_real*0.434)*dt,min_erates,np.infty),Calphahats_real,reset


from evspy.utils.loadsteps import LoadSteps
from evspy.multi_stage_model.decoupled_fdm import distorted_isotache_model
import numpy as np

def test_decoupled_consolidation_IL():
    load = 'test'
    load = LoadSteps()
    load.add_load_step(duration = 1e8,load = 5,rate = 1e-8,type_test = 'IL',
                        cv=7)
    model = distorted_isotache_model(load,sigma0=1,H=0.1,dimt=1000,e_init=1)
    model.initialize_decoupled()
    model.run_iterations()
    assert model.sigma[-1] == 5

def test_decoupled_consolidation_CRS():
    load=LoadSteps()
    load.add_load_step(duration = 1e8,
                        load = 5,
                        rate = 1e-8,
                        type_test = 'CRS',
                        cv=7)
    model = distorted_isotache_model(load,sigma0=1,H=0.1,dimt=1500,e_init=1)
    model.initialize_decoupled()
    model.run_iterations()
    assert np.round(model.sigma[-1],1) == 5
    assert np.round(np.log10(-model.erate[-1]),1) == -8


import numpy as np
import pandas as pd
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
from pynverse import inversefunc
#import lmfit
from scipy.special import lambertw


from evspy.utils.helper_functions import line_intersection, get_intersection # Refactoring needed for get_intersection?
from evspy.single_stage_model.creep_swelling_model import beta3_fit, beta2_fit, power_law, beta_nash
# Strain rate relations

def Calpha_from_erate(erate,eref=1e-5,Cc=0.2,Cr=0.2/5,Calpha=0.2*0.04,beta2=3,beta3=19):
    K=(Cc-Cr)/Calpha
    try:
        erate.shape[1]
        erate=erate[0]
    except:
        pass
    R=erate/eref
    Z=-2*beta2/K*np.log(R)+1
    OCR_x=lambertw(np.exp(Z))**(1/beta2)
    return 2/(np.array([np.float(i) for i in OCR_x])**beta2+1)

def mitsr_strain_rate(OCR,beta2,Cc,Cr,Calpha):
    return (np.sqrt(1/OCR)*np.exp((1-(OCR)**beta2)/(2*beta2)))**((Cc-Cr)/Calpha)

def log_mitsr_strain_rate(OCR,beta2,Cc,Cr,Calpha):
    return np.log10((np.sqrt(1/OCR)*np.exp((1-(OCR)**beta2)/(2*beta2)))**((Cc-Cr)/Calpha))

def get_sigref(sigma,e,Cc,Cr,e0,loglog=False):
    if loglog:
        rhor=Cr/e0
        rhoc=Cc/e0
        X,Y=line_intersection(((np.log(sigma),np.log(e)),(5,np.log(e)-rhor*np.log(np.exp(5)/sigma))),((0,np.log(e0)),(5,np.log(e0)-rhoc*5)))
        return np.exp(X),np.exp(Y)
    else:
        X,Y=line_intersection(((np.log10(sigma),e),(5,e-Cr*np.log10(10**5/sigma))),((0,e0),(5,e0-Cc*5)))
        return 10**X,Y

def create_isotache(erateref = 1e-5,beta2 = 4,beta3 = 25,Cc = 0.2,Cr = 0.2 /5,CalphaNC = 0.2 * 0.04,sigp=1.,e0=1.5,isotache=False,power_law=False,ref_func='semilogx',m=1,beta_nash=False,m0=-1.22,b0=-1.14,param_nash=[0.02952926,0.02696060,9.03771812],dsig=0.05,Calpha_OCRref=True,beta2beta3=False):
    #lookup=np.clip(2/(inversefunc(log_mitsr_strain_rate,y_values=np.arange(-100,0,0.1),domain=0.00001,args=(beta2,Cc,Cr,CalphaNC))**beta2+1),0.0,1)*CalphaNC
    if isotache:
        sigp=1
    sigrange = 10**np.arange(-1,3,dsig)
    sigrangeNC =  10**np.arange(-3,10,dsig)
    sig_start = sigrangeNC[0]
    OCR = np.clip((sigp)/(sigrange),1,np.infty)
    if ref_func=='loglog':
        rhoc=Cc/e0
        rhor=Cr/e0
        eNC = np.exp(np.log(e0) - rhoc*np.log(sigrangeNC))
        eUC0 = np.exp(np.log(e0) +(sigrange<sigp)*(- rhoc*np.log(sigp) - (-rhor*np.log(sigp)+rhor*np.log(sigrange))) - (sigrange>=sigp)*rhoc*np.log(sigrange))
    elif ref_func =='semilogx':
        eNC = e0 - Cc*np.log10(sigrangeNC)
        eUC0 = e0 +(sigrange<sigp)*(- Cc*np.log10(sigp) - (-Cr*np.log10(sigp)+Cr*np.log10(sigrange))) - (sigrange>=sigp)*Cc*np.log10(sigrange)
    elif ref_func=='flexible':
        slope_flex = Cc/m/np.log10((sig_start**-1))**(m-1)
        eNC = e0 -slope_flex*np.log10(sigrangeNC*(sig_start**-1))**m+slope_flex*np.log10((sig_start**-1))**m
        eUC0 = e0 +(sigrange<sigp)*((-slope_flex*np.log10(sigp*(sig_start**-1))**m+slope_flex*np.log10((sig_start**-1))**m) - 
                                    (-Cr*np.log10(sigp)+Cr*np.log10(sigrange))) - (sigrange>=sigp)*(slope_flex*np.log10(sigrange*(sig_start**-1))**m-slope_flex*np.log10((sig_start**-1))**m)
    if isotache:
        eratei = mitsr_strain_rate(OCR,beta2,Cc,Cr,CalphaNC)*erateref
    else:
        eratei = beta3_fit(beta3,OCR)*erateref
    if power_law:
        Calphac = np.clip(power_law(np.clip(OCR,1.01,np.infty),m0,b0),0,1)*CalphaNC
    elif beta_nash:
        Calphac = np.clip(beta_nash(OCR,param_nash[0],param_nash[1],param_nash[2]),0,1)*CalphaNC
    else:
        if isotache:
            Calphac = np.clip(beta2_fit(beta2,OCR),0,1)*CalphaNC
        else:
            if beta2beta3:
                Calphac = np.clip(2/((eratei/erateref)**(-beta2/beta3)+1),0.1,1)*CalphaNC
            else:
                #Calphac = np.clip(beta2_fit(beta2,OCR),0,1)*CalphaNC
                Calphac = CalphaNC*Calpha_from_erate(eratei,eref=erateref,Cc=Cc,Cr=Cr,Calpha=CalphaNC,beta2=beta2,beta3=beta3)
            
    df=pd.DataFrame()
    df['sigma']=sigrange
    df['e']=eUC0
    eUC=eUC0
    df['erate']=eratei
    sigp0 = np.clip(sigp,sigrange,np.infty)
    OCR = np.clip((sigp)/(sigrange),1,np.infty)
    sigp = sigp0
    for i in range(500):
        deltae = Calphac/6 #0.25 log cycle down
        eUC = eUC-deltae
        eratei=10**(np.log10(eratei)-deltae/Calphac)
        #eratei = np.clip(10**(np.log10(eratei)-deltae/Calphac),1e-1000,1)
        if Calpha_OCRref:
            sigp = get_intersection(sigrange,eUC,sigrangeNC,eNC,Cr)[0]#(sigp+deltae/(Cc-Cr))
        OCR = sigp / sigrange
        if power_law:
            Calphac = np.clip(power_law(np.clip(OCR,1.01,np.infty),m0,b0),0,1)*CalphaNC
        elif beta_nash:
            Calphac = np.clip(beta_nash(OCR,param_nash[0],param_nash[1],param_nash[2]),0,1)*CalphaNC #param_nash=[betamin,b0,a]
        else:
            if isotache:
                Calphac = np.clip(beta2_fit(beta2,OCR),0,1)*CalphaNC
            else:
                if beta2beta3:
                    Calphac = np.clip(2/((eratei/erateref)**(-beta2/beta3)+1),0.1,1)*CalphaNC
                else:
                    #Calphac = np.clip(beta2_fit(beta2,OCR),0,1)*CalphaNC
                    Calphac = CalphaNC*Calpha_from_erate(eratei,eref=erateref,Cc=Cc,Cr=Cr,Calpha=CalphaNC,beta2=beta2,beta3=beta3)
                
        dftemp=pd.DataFrame()
        dftemp['sigma']=sigrange
        dftemp['e']=eUC
        dftemp['erate']=eratei
        df=df.append(dftemp)
    eUC = eUC0
    sigp = sigp0
    OCR = sigp / sigrange
    if isotache:
        eratei = mitsr_strain_rate(OCR,beta2,Cc,Cr,CalphaNC)*erateref
    else:
        eratei = beta3_fit(beta3,OCR)*erateref
    if power_law:
        Calphac = np.clip(power_law(np.clip(OCR,1.01,np.infty),m0,b0),0,1)*CalphaNC
    elif beta_nash:
        Calphac = np.clip(beta_nash(OCR,param_nash[0],param_nash[1],param_nash[2]),0,1)*CalphaNC #param_nash=[betamin,b0,a]
    else:
        if isotache:
            Calphac = np.clip(beta2_fit(beta2,OCR),0,1)*CalphaNC
        else:
            if beta2beta3:
                Calphac = np.clip(2/((eratei/erateref)**(-beta2/beta3)+1),0.1,1)*CalphaNC
            else:
                #Calphac = np.clip(beta2_fit(beta2,OCR),0,1)*CalphaNC
                Calphac = CalphaNC*Calpha_from_erate(eratei,eref=erateref,Cc=Cc,Cr=Cr,Calpha=CalphaNC,beta2=beta2,beta3=beta3)
    for i in range(500):
        deltae = -Calphac/6 #0.25 log cycle down
        eUC = eUC-deltae
        eratei=10**(np.log10(eratei)-deltae/Calphac)
        #eratei = np.clip(10**(np.log10(eratei)-deltae/Calphac),1e-1000,1)
        if Calpha_OCRref:
            #sigp = get_intersection(sigrange,eUC,sigrangeNC,eNC,Cr)[0]#
            sigp = (sigp+deltae/(Cc-Cr))
        OCR = sigp / sigrange
        if power_law:
            Calphac = np.clip(power_law(np.clip(OCR,1.01,np.infty),m0,b0),0,1)*CalphaNC
        elif beta_nash:
            Calphac = np.clip(beta_nash(OCR,param_nash[0],param_nash[1],param_nash[2]),0,1)*CalphaNC #param_nash=[betamin,b0,a]
        else:
            if isotache:
                Calphac = np.clip(beta2_fit(beta2,OCR),0,1)*CalphaNC
            else:
                if beta2beta3:
                    Calphac = np.clip(2/((eratei/erateref)**(-beta2/beta3)+1),0.1,1)*CalphaNC
                else:
                    #Calphac = np.clip(beta2_fit(beta2,OCR),0,1)*CalphaNC
                    Calphac = CalphaNC*Calpha_from_erate(eratei,eref=erateref,Cc=Cc,Cr=Cr,Calpha=CalphaNC,beta2=beta2,beta3=beta3)
                

        dftemp=pd.DataFrame()
        dftemp['sigma']=sigrange
        dftemp['e']=eUC
        dftemp['erate']=eratei
        df=df.append(dftemp) 
    return df,eUC,eUC0,eNC,sigrange,sigrangeNC

def calc_pos_isotache(sigma,e,erateref = 1e-5,beta2 = 4,beta3 = 25,Cc = 0.2,Cr = 0.2 /5,CalphaNC = 0.2 * 0.04,sigp=100.,e0=1.5,isotache=False,power_law=False,ref_func='semilogx', Calpha_OCRref=True,m=1,m0=-1.22,b0=-1.14,dsig=0.05,steps=300,beta_nash=False,param_nash=[0.02952926,0.02696060,9.03771812],lookup=[],beta2beta3=False):
    if len(lookup)==0:
        lookup=np.clip(2/(inversefunc(log_mitsr_strain_rate,y_values=np.arange(-100,0,0.1),domain=0.00001,args=(beta2,Cc,Cr,CalphaNC))**beta2+1),0.0,1)*CalphaNC
    
        
def make_contour(df,eUC,eUC0,eNC,sigrange,sigrangeNC,num=20,figsize=(8,5),vmin=-65,vmax=-5,colorbar=False):
    sigi=sigrange
    ei=10**np.arange(np.log10(0.5),np.log10(3),0.0005)
    erateii=griddata((df['sigma'],df['e']),np.log10(df['erate']),xi=(sigi[None,:],ei[:,None]))
    for i in range(len(ei)):
        for j in range(len(sigi)):
            if ei[i]>eUC[sigrange==sigrange[j]][0]:
                erateii[i,j]=np.nan
    
    plt.figure(figsize=figsize)
    ax=plt.subplot(111)
    #ax.annotate(r'$C_c$',xy=(10,1.75),xytext=(50,1.55))
    #ax.annotate(r'reference',xy=(10,1.75),xytext=(15,1.73))
    #ax.annotate(r'isotache',xy=(10,1.75),xytext=(22,1.69))
    #ax.annotate(r'$C_r$',xy=(4,1.6),xytext=(4,1.7))
    cs=plt.contour(sigi,ei,np.clip(erateii,vmin,vmax),num,vmax=vmax,vmin=vmin,cmap='jet')
    plt.semilogx(sigrangeNC,eNC,'k--')
    plt.plot(sigrange,eUC0,'k')
    if colorbar:
        plt.colorbar(label=r'$log_{10}(\dot{e}_c)$')
    plt.ylim(1.4,1.8)
    plt.xlim(2,100)
    plt.tight_layout()
    plt.xlabel("Stress, $\sigma_v'$ (kPa)"); plt.ylabel('Void ratio, $e$')#
    return cs    
    

from functools import wraps
import inspect

def instanceVariables(func):
    def returnFunc(*args, **kwargs):
        selfVar = args[0]

        argSpec = inspect.getargspec(func)
        argumentNames = argSpec[0][1:]
        defaults = argSpec[3]
        if defaults is not None:
            defaultArgDict = dict(zip(reversed(argumentNames), reversed(defaults)))
            selfVar.__dict__.update(defaultArgDict)

        argDict = dict(zip(argumentNames, args[1:]))
        selfVar.__dict__.update(argDict)


        validKeywords = set(kwargs) & set(argumentNames)
        kwargDict = {k: kwargs[k] for k in validKeywords}
        selfVar.__dict__.update(kwargDict)

        func(*args, **kwargs)

    return returnFunc

class base_isotache_model:

    @instanceVariables
    def __init__(self,load,sigma0,H,dimt=8000,erateref = 1e-5,erateinits=1e-10,
                    beta2 = 4,beta3 = 25,Cc = 0.2, Cr = 0.2 /5,CalphaNC = 0.2 * 0.04,
                    Cv=7,sigp=100.,e0=1.5,isotache=False,use_power_law=False,ref_func='semilogx', 
                    e_init=1.9,Calpha_OCRref=True,m0=-1.22,b0=-1.14,
                    m1=0.88,b1=-0.3,b2=-5.,m2=2.,dsig=0.05,steps=300,
                    use_beta_nash=False,param_nash=[0.02952926,0.02696060,9.03771812],
                    beta2beta3=False):
        self.eratec=np.zeros((dimt)); self.erates=np.zeros((dimt)); self.erate_e=np.zeros((dimt))
        self.e=np.zeros((dimt)); self.erate=np.zeros((dimt)); self.ee=np.zeros((dimt))
        self.OCR_real=np.zeros((dimt)); self.OCR_ref=np.zeros((dimt)); self.OCR=np.zeros((dimt))
        self.sigp_update=np.zeros((dimt)); self.sigpref=np.zeros((self.dimt))
        self.lookup = np.clip(2/(inversefunc(log_mitsr_strain_rate,y_values=np.arange(-100,0,0.1),
                                                domain=0.00001,args=(self.beta2,self.Cc,self.Cr,self.CalphaNC))**beta2+1),0.0,1)*self.CalphaNC

        self.CalphaCc = self.CalphaNC/self.Cc

    def get_sigpref(self,sigma,e,sigpref,initiate=False):
        if (self.ref_func =='semilogx')&(initiate==False):
            sigpref = 10**(np.log10(sigpref[0])-(e[1]-e[0])/(self.Cc-self.Cr)-self.Cr/(self.Cc-self.Cr)*np.log10(sigma[1]/sigma[0]))
        else:      
            sigpref = get_intersection(sigma,e,self.sigrangeNC,self.eNC,self.Cr)[0][0]
        return sigpref
        
    def initialize_reference_curve(self):
        self.sigrangeNC =  10**np.arange(-2,10,self.dsig)
        if self.ref_func=='loglog':
            self.rhoc=self.Cc/self.e0
            self.rhor=self.Cr/self.e0
            self.eNC = np.exp(np.log(self.e0) - self.rhoc*np.log(self.sigrangeNC))
        elif self.ref_func =='semilogx':
            self.eNC = self.e0 - self.Cc*np.log10(self.sigrangeNC)
                   
    def calc_pos_isotache(self,sigma,e,sigp,steps=200):
        sigp = np.clip(sigp,sigma,np.infty)
        OCR = np.clip((sigp)/(sigma),1,np.infty)
        if self.ref_func=='loglog':
            self.eUC0 = np.exp(np.log(self.e0) +(sigma<sigp)*(- self.rhoc*np.log(sigp) - (-self.rhor*np.log(sigp)+self.rhor*np.log(sigma))) - (sigma>=sigp)*self.rhoc*np.log(sigma))
        elif self.ref_func =='semilogx':
            self.eUC0 = self.e0 +(sigma<sigp)*(- self.Cc*np.log10(sigp) - (-self.Cr*np.log10(sigp)+self.Cr*np.log10(sigma))) - (sigma>=sigp)*self.Cc*np.log10(sigma)

        if self.isotache:
            eratei = mitsr_strain_rate(OCR,self.beta2,self.Cc,self.Cr,self.CalphaNC)*self.erateref
        else:
            eratei = beta3_fit(self.beta3,OCR)*self.erateref
        if self.use_power_law:
            Calphac = np.clip(power_law(np.clip(OCR,1.01,np.infty),self.m0,self.b0),0,1)*self.CalphaNC
        elif self.use_beta_nash:
            Calphac = np.clip(beta_nash(OCR,self.param_nash[0],self.param_nash[1],self.param_nash[2]),0,1)*self.CalphaNC #param_nash=[betamin,b0,a]
        else:
            if self.isotache:
                Calphac = np.clip(beta2_fit(self.beta2,OCR),0,1)*self.CalphaNC
            else:
                if self.beta2beta3:
                    Calphac = np.clip(2/((eratei/self.erateref)**(-self.beta2/self.beta3)+1),0.1,1)*self.CalphaNC
                else:
                    #Calphac = np.clip(fit_func.beta2_fit(beta2,OCR),0,1)*CalphaNC
                    Calphac = self.CalphaNC*Calpha_from_erate(np.array([eratei]),eref=self.erateref,Cc=self.Cc,Cr=self.Cr,
                                                Calpha=self.CalphaNC,beta2=self.beta2,beta3=self.beta3)[0]
                    #Calphac = np.interp(eratei/erateref,10**np.arange(-100,0,0.1),lookup)
                    
        deltae_target=e-self.eUC0
        eUC=self.eUC0
        eratei0=eratei
        for i in range(steps):
            deltae = -deltae_target/steps 
            eUC = eUC-deltae
            eratei = 10**(np.log10(eratei)-deltae/Calphac)
            #sigp =(sigp+deltae/(Cc-Cr))
            if self.Calpha_OCRref:
                if self.ref_func =='semilogx':
                    sigp = sigp+deltae/(self.Cc-self.Cr)*sigp
                elif self.ref_func == 'loglog':
                    sigp = sigp+np.log10((eUC+deltae)/eUC)/(self.rhoc-self.rhor)*sigp
                else:
                    sigp = get_intersection(np.array([sigma]),np.array([eUC]),self.sigrangeNC,self.eNC,self.Cr)[0][0]#
            #print(Calphac[1])
            OCR = sigp / sigma
            if self.use_power_law:
                Calphac = np.clip(power_law(np.clip(OCR,1.01,np.infty),self.m0,self.b0),0,1)*self.CalphaNC
            elif self.use_beta_nash:
                Calphac = np.clip(beta_nash(OCR,self.param_nash[0],self.param_nash[1],self.param_nash[2]),0,1)*self.CalphaNC #param_nash=[betamin,b0,a]
            else:
                if self.isotache:
                    Calphac = np.clip(beta2_fit(self.beta2,OCR),0,1)*self.CalphaNC
                else:
                    if self.beta2beta3:
                        Calphac = np.clip(2/((eratei/self.erateref)**(-self.beta2/self.beta3)+1),0.1,1)*self.CalphaNC
                    else:
                        #Calphac = np.clip(fit_func.beta2_fit(beta2,OCR),0,1)*CalphaNC
                        Calphac = self.CalphaNC*Calpha_from_erate(np.array([eratei]),eref=self.erateref,Cc=self.Cc,Cr=self.Cr,Calpha=self.CalphaNC,beta2=self.beta2,beta3=self.beta3)[0]
                        #Calphac = np.interp(eratei/erateref,10**np.arange(-100,0,0.1),lookup)
        return eratei




import evspy.single_stage_model.creep_swelling_model as cs

class decoupled_cs_model:
    def __init__(self, Calphahatc, Calphahats,  erateci, eratesi, Cc, Cr, CalphaNC, OCR_initial,OCR_final, sigma0 = 100, sigmap = 100,
                 dimt=100000,dtfactor=1e-7,Cv=5/3600/24/365,H=0.01,targettime=1e10, swelling_isotache = False,
                 nonlinear_Calphahat=True,sigmaref=100,beta2=4,beta3=19,hypA=False,
                model_type = 'load_controlled'):
        self.Calphahatc = Calphahatc
        self.Calphahats = Calphahats
        self.CalphaNC = CalphaNC
        self.erateci = erateci
        self.eratesi = eratesi
        self.Cc = Cc
        self.Cr = Cr
        self.OCR_initial = OCR_initial
        self.OCR_final = OCR_final
        self.dimt = dimt
        self.dtfactor = dtfactor
        self.Cv = Cv
        self.H = H
        self.targettime = targettime
        self.nonlinear_Calphahat = nonlinear_Calphahat
        self.sigmaref = sigmaref
        self.beta2 = beta2
        self.beta3 = beta3
        self.hypA = hypA
        self.model_type = model_type
        self.g = 9.81
        self.sigma0 = sigma0
        self.swelling_isotache = swelling_isotache
        self.sigmap = sigmap
        
        if model_type == 'load_controlled':
            res = cs.C_S_model_with_load(self.Calphahatc, self.Calphahats,  self.erateci, self.eratesi, self.Cr,self.Cc, 
                                        self.OCR_initial,self.OCR_final,dimt=self.dimt,dtfactor = self.dtfactor,Cv = self.Cv,H = self.H,targettime = self.targettime,
                                            nonlinear_Calphahat = self.nonlinear_Calphahat, sigmaref = self.sigmaref,beta2 = self.beta2,beta3 = self.beta3,hypA = self.hypA)
        elif model_type == 'relaxation':
            res = cs.C_S_relaxation(self.Calphahatc, self.Calphahats,  self.erateci, self.eratesi, self.Cr,self.Cc,self.sigma0,dimt=self.dimt,dtfactor = self.dtfactor,
                                             targettime = self.targettime,nonlinear_Calphahat = self.nonlinear_Calphahat, swelling_isotache=self.swelling_isotache,
                                             OCR = self.OCR_final, beta2 = self.beta2, beta3 = self.beta3)
        elif model_type == 'rate_controlled':
            res = cs.C_S_CRS(self.Cc,self.Cr,self.sigmap,self.erateci,self.CalphaNC,erateref = self.erateref, Cc_reduction=2.3,
                                Cc_f_OCR=False,erates0=-1,erate = 1,sigma0=265,sigmaf=160,b2=-5.5,m2=2.5,beta2=self.beta2,beta3=self.beta3,estart=1,m1=0.88,b1=-1.4,dtfactor=1)

        self.time = res['time']
        self.e = res['e']
        self.erate = res['erate']
        self.erate_c = res['erate_c']
        self.erate_s = res['erate_s']
        self.Calphahatc = res['Calphahatc']
        if model_type == 'load_controlled':
            self.OCR = res['OCR']
            self.OCR_ref = res['OCR_ref']
            self.erate_e = res['erate_e']
        if model_type == 'relaxation':
            self.sigma = res['sigma']
        



import pandas as pd
import numpy as np


class LoadSteps:
    def __init__(self):
        self.load_steps = [0]
        self.duration = [0]
        self.start_time = [0]
        self.end_time = [0]
        self.load = [0]
        self.type = ['IL']
        self.rate = [0]
        self.cv = [1]
        self.loaddf = pd.DataFrame(columns=['Load'])
    def add_load_step(self, duration, load, rate, type_test,cv):
        self.load_steps.append(len(self.load_steps))
        self.duration.append(duration)
        self.start_time.append(self.end_time[-1])
        self.end_time.append(self.start_time[-1]+duration)
        self.load.append(load)
        self.rate.append(rate)
        self.type.append(type_test)
        self.cv.append(cv)
        self.loaddf.loc[self.start_time[-1]]=self.load[-1]
        self.loaddf.loc[self.end_time[-1]-0.1]=self.load[-1]
    def change_load_step(self,load_step,duration,load, rate, type_test):
        self.duration[load_step]=duration
        self.load[load_step]=load
        self.rate[load_step]=rate
        self.type[load_step]=type_test
        self.end_time[load_step]=self.start_time[load_step]+duration
        self.loaddf = pd.DataFrame(columns=['Load'])
        for i in range(1,self.load_steps[-1]+1):
            self.start_time[i]=self.end_time[i-1]
            self.end_time[i]=self.start_time[i]+self.duration[i]
            self.loaddf.loc[self.start_time[i]]=self.load[i]
            self.loaddf.loc[self.end_time[i]-0.1]=self.load[i]
    def delete_load_step(self,load_step):
        self.load_steps.remove(load_step)
        self.duration.remove(self.duration[load_step])
        self.load.remove(self.load[load_step])
        for i in range(load_step,self.load_steps[-1]):
            self.load_steps[i]=self.load_steps[i-1]+1
            self.start_time[i]=self.end_time[i-1]
            self.end_time[i]=self.start_time[i]+self.duration[i]
        self.loaddf = pd.DataFrame(columns=['Load'])
        for i in range(1,self.load_steps[-1]+1):
            self.loaddf.loc[self.start_time[i]]=self.load[i]
            self.loaddf.loc[self.end_time[i]-0.1]=self.load[i]  

def geo_mean(iterable):
    a = np.array(iterable)
    return a.prod()**(1.0/len(a))

def log_mean(x,y):
    return np.mean(np.interp(10**np.linspace(np.log10(100),np.log10(np.max(x)),100),x,y))

def identify_all_stages(oed_data):
    stages = list(set(oed_data['step_id']))
    stage_info={}
    for i in stages:
        stage_info[i]={}
        stage_data=oed_data[oed_data['step_id']==i]
        stage_info[i]['avg_void_ratio']=log_mean(stage_data['time'],stage_data['void_ratio'])
        stage_info[i]['avg_load']=np.average(stage_data['load'])
        stage_info[i]['initial_void_ratio']=stage_data['void_ratio'].iloc[0]
        stage_info[i]['final_void_ratio']=stage_data['void_ratio'].iloc[-1]
        stage_info[i]['sample_height']=stage_data['sample_height'].iloc[0]
        stage_info[i]['duration']=stage_data['time'].iloc[-1]-stage_data['time'].iloc[0]
        if i==1:
            if np.std(stage_data['void_ratio_rate'].iloc[5:-5])/np.nanmean(stage_data['void_ratio_rate'])<0.1:
                stage_info[i]['type']='CRS'
                stage_info[i]['strain_rate']=np.nanmedian(stage_data['void_ratio_rate'])
            else:
                stage_info[i]['type']='IL'
                stage_info[i]['strain_rate']=np.nan
        elif np.std(stage_data['void_ratio'])<1e-5:
            stage_info[i]['type']='R'
            stage_info[i]['strain_rate']=0
        elif np.abs(oed_data[oed_data['step_id']==i-1]['load'].iloc[-1]-stage_data['load'].iloc[-1])<0.1:
            stage_info[i]['type']='CL'
            stage_info[i]['strain_rate']=np.nan
        elif  np.abs((np.nanmean(stage_data['void_ratio_rate'].iloc[np.int(len(stage_data)/2):-1])-np.nanmean(stage_data['void_ratio_rate'].iloc[1:-np.int(len(stage_data)/2)]))/np.nanmean(stage_data['void_ratio_rate'].iloc[1:-1]))<0.05:
            stage_info[i]['type']='CRS'
            #print(np.nanmean(stage_data['void_ratio_rate'].iloc[int(len(stage_data)/3):int(len(stage_data)*2/3)]))
            stage_info[i]['strain_rate']=np.nanmean(stage_data['void_ratio_rate'].iloc[int(len(stage_data)/4):int(len(stage_data)*3/4)])
        else:
            stage_info[i]['type']='IL'
            stage_info[i]['strain_rate']=np.nan
        stage_info[i]['initial_OCR']=np.max(oed_data[oed_data['step_id']<i]['load'])/stage_data['load'].iloc[0]#oed_data[oed_data['step_id']==np.clip(i-1,1,np.infty)]['load'].iloc[-1]
        stage_info[i]['final_OCR']=np.max(oed_data[oed_data['step_id']<i]['load'])/stage_data['load'].iloc[-1]
        stage_info[i]['initial_load']=stage_data['load'].iloc[0]
        stage_info[i]['final_load']=stage_data['load'].iloc[-1]
        stage_info[i]['load_change']=stage_data['load'].iloc[-1]-stage_data['load'].iloc[0]
    return stage_info


import numpy as np

def Consol_Terzaghi_Uavg_vertical(Cv,H,t=-1,targettime=1000,dimt=1000,time=-1,constant_dt_time=np.infty,dtmax=1e3):
    '''
    Parameters
    ----------
    Cv : float
        coefficient of vertical consolidation in m/s**2
    '''
    if dimt==-1:
        time=time
    else:
        if constant_dt_time>targettime:
            time=10**np.linspace(0,np.log10(targettime),dimt)
        else:
            dimt2 = np.int((targettime-constant_dt_time)/dtmax)
            time=np.append(10**np.linspace(0,np.log10(constant_dt_time),dimt),np.linspace(constant_dt_time+0.01,targettime,dimt2))
    if t>0:
        Tv=t*Cv/((H)**2)
    else:
        Tv=time*Cv/((H)**2)
    Tv=np.clip(Tv,0,10)
    UavgAnalV=np.sqrt(4*Tv/np.pi)/((1+(4*Tv/np.pi)**2.8)**0.179)
    return time,UavgAnalV


import numpy as np
from scipy.optimize import fsolve
from scipy.interpolate import griddata
import matplotlib.pyplot as plt
from scipy.special import lambertw
g=9.81; gammaw=1*g


def C_S_model_with_load(Calphahatc,Calphahats,erateci,eratesi,Cr,Cc,OCR_initial,OCR_final,dimt=100000,dtfactor=1e-7,Cv=5/3600/24/365,H=0.01,targettime=1e10,nonlinear_Calphahat=False,sigmaref=100,beta2=4,beta3=19,hypA=False):
    '''
    Parameters
    ----------
    Calphahatc : float
        secondary compression index (creep) after unloading at OCR_final
    
    Returns
    -------
    
    '''
    CcCrRatio = Cc/Cr
    time=np.zeros((dimt))
    eratec=np.zeros((dimt)); erates=np.zeros((dimt)); erate_e=np.zeros((dimt))
    e=np.zeros((dimt)); erate=np.zeros((dimt)); ee=np.zeros((dimt))
    eratec[0]=erateci
    erates[0]=0#eratesi
    erate[0]=eratec[0]-erates[0]
    dt=1e-20
    OCR_ref=np.zeros((dimt)); OCR=np.zeros((dimt))
    OCR_ref[0]=OCR_initial
    sigma=np.zeros((dimt))
    sigma[0]=sigmaref
    sigp_ref=np.zeros((dimt))
    sigp_ref[0]=sigmaref*OCR_initial
    Calphahatc_real=np.zeros((dimt))
    Calphahats_real=np.zeros((dimt))
    erateref=erateci/beta3_fit(beta3,OCR_final)
    t=0
    if nonlinear_Calphahat==True:
        Calphahatc_real[0]=Calphahatc#*ratio_Calphacinit(OCR_initial,beta2)/ratio_Calphacinit(OCR_final,beta2)
        Calphahats_real[0]=Calphahatc*ratio_Calphasinit(OCR_initial)/ratio_Calphasinit(OCR_final)
    else:
        Calphahatc_real[0]=Calphahatc
        Calphahats_real[0]=Calphahats
    #start=False
    i=0
    failsafe=0
    dtmult = 1.01
    dtfactorinit=dtfactor
    while (time[t]<targettime) and (t<dimt-1) and (dt>0):
        if  (dt>0)&(eratec[t]>-0.01)&(erates[t]>-0.01)&(np.abs(eratec[t]*dt)<1e-2)&(np.abs(erates[t]*dt)<1e-2):
            dtfactor=dtfactor * dtmult
            t+=1
            #print(dt)
            time[t]=time[t-1]+dt
            time2,UavgAnal=Consol_Terzaghi_Uavg_vertical(Cv,H,t=time[t],dimt=-1)
            sigma[t]=sigmaref-(sigmaref-(OCR_initial/OCR_final*sigmaref))*np.clip(UavgAnal,0,1)
            OCR_ref[t]=sigp_ref[t-1]/sigma[t]
            OCR[t]=sigmaref*OCR_initial/sigma[t]
            if (OCR[t]<OCR_final)&(OCR[t]>1.):
                erates[t-1]=erates[t-1]*10**(5*np.log10((OCR[t]-1)/(OCR[t-1]-1))) # Viscous swelling develops and reduces at the same time while unloading
            if t==1:
                erates[t-1]=eratesi*10**(5*np.log10(np.clip((OCR[t]-1)/(OCR_final-1),0,np.infty)))
            if (OCR[t]<1.05)&hypA:
                erate_e[t]= Cc*0.434*(sigma[t]-sigma[t-1])/dt/sigma[t]
            else:
                erate_e[t]= Cr*0.434*(sigma[t]-sigma[t-1])/dt/sigma[t]
            ee[t]=ee[t-1]+erate_e[t]*dt 
            if hypA:
                e[t]=e[t-1]+(eratec[t-1]-erates[t-1])*dt*(np.abs(OCR[t]-OCR_final)<0.1)+(ee[t]-ee[t-1]) #
            else:
                e[t]=e[t-1]+(eratec[t-1]-erates[t-1])*dt+(ee[t]-ee[t-1])

            if (OCR_final < OCR_initial) and not hypA:
                eratec[t]=eratec[t-1]-eratec[t-1]*(eratec[t-1]-erates[t-1])*1/(Calphahatc_real[t-1]*0.434)*dt+eratec[t-1]*erate_e[t]*(Cc-Cr)/Cr*1/(Calphahatc_real[t-1]*0.434)*dt 
            else:
                eratec[t]=eratec[t-1]-eratec[t-1]*(eratec[t-1]-erates[t-1])*1/(Calphahatc_real[t-1]*0.434)*dt
              
            erates[t]=erates[t-1]-erates[t-1]*erates[t-1]*1/(Calphahats_real[t-1]*0.434)*dt 
            erate[t]=(e[t]-e[t-1])/dt#
            sigp_ref[t]=sigp_ref[t-1]
            sigp_ref[t]=sigp_ref[t]+sigp_ref[t]*(eratec[t]-erates[t])/(Cr*CcCrRatio-Cr)/0.434*dt

            # Calphac evolution related to eratec instead of OCRref
               
            if nonlinear_Calphahat==True:
                Calphahats_real[t]=ratio_Calphasinit(sigp_ref[t]/sigma[t])*(Calphahats/ratio_Calphasinit(OCR_final))
                Calphahatc_real[t]=Calphahatc*ratio_Calphac_f_rate(eratec[t]/erateref,beta2,beta3)/ratio_Calphac_f_rate(erateci/erateref,beta2,beta3)#
            else:
                Calphahatc_real[t]=Calphahatc
                Calphahats_real[t]=Calphahats
            dt=np.clip(1/np.abs(eratec[t]-erates[t])*dtfactor,1e-100,time[t]/2)
        else:
            t = 2
            failsafe += failsafe
            dtfactor = dtfactorinit / 10
            dtmult = 1+(dtmult-1) / 2
            dtfactorinit = dtfactorinit/10
            dt=1e-100
    if failsafe > 9:
        print(Calphahatc,Calphahats,erateci,eratesi,Cr)
    return {'time':time[:t],'e':e[:t],'erate':erate[:t],'erate_c':eratec[:t],'erate_s':erates[:t],'Calphahatc':Calphahatc_real[:t],'OCR':OCR[:t],'OCR_ref':OCR_ref[:t],'erate_e':erate_e[:t]}


def Calpha_from_erate(erate,eref=1e-5,Cc=0.2,Cr=0.2/5,Calpha=0.2*0.04,beta2=3,beta3=19):
    K=(Cc-Cr)/Calpha
    R=erate/eref
    Z=-2*beta2/K*np.log(R)+1
    OCR_x=lambertw(np.exp(Z))**(1/beta2)
    return 2/(OCR_x**beta2+1)

def C_S_model_with_ocr(Calphahatc,Calphahats,erateci,eratesi,Cr,OCRtarget,Cc,dimt=100000,dtfactor=1e-7,Cv=5/3600/24/365,H=0.01,targettime=1e10,creep_coupled=False,sigmaref=100,beta2=3,beta3=19):
    with np.errstate(divide='ignore', invalid='ignore'):
        CcCrRatio=Cc/Cr
        dtfactorinit=dtfactor
        time=np.zeros((dimt))
        eratec=np.zeros((dimt)); erates=np.zeros((dimt))
        e=np.zeros((dimt)); erate=np.zeros((dimt)); ee=np.zeros((dimt))
        eratec[0]=erateci
        erates[0]=0
        erate[0]=eratec[0]-erates[0]
        dt=1e-20
        OCR=np.zeros((dimt))
        OCR[0]=1
        sigp_ref=np.zeros((dimt))
        sigp_ref[0]=sigmaref
        Calphahatc_real=np.zeros((dimt))
        t=0
        erateref=erateci/beta3_fit(beta3,OCRtarget)
        if creep_coupled==True:
            Calphahatc_real[0]=Calphahatc
        else:
            Calphahatc_real[0]=Calphahatc
        #start=False
        i=0
        failsafe=0
        dtmult = 1.01
        while (time[t]<targettime*1.5) and (t<dimt-1) and (failsafe<10):
            if  (dt>0)&(eratec[t]>-0.01)&(erates[t]>-0.01)&(np.abs(eratec[t]*dt)<1e-2)&(np.abs(erates[t]*dt)<1e-2):
                dtfactor=dtfactor * dtmult
                t+=1
                #print(dt)
                time[t]=time[t-1]+dt
                time2,UavgAnal=Consol_Terzaghi_Uavg_vertical(Cv,H,t=time[t])
                OCR[t]=1/(1-np.clip(UavgAnal,0,1)+1/OCRtarget*np.clip(UavgAnal,0,1))
                if OCR[t]<OCRtarget:
                    erates[t-1]=erates[t-1]*10**(5*np.log10(np.clip(OCR[t]-1,0.0000000000001,np.infty)/(OCR[t-1]-1))) # Viscous swelling develops and reduces at the same time while unloading ==> TO BE CHECKED EXPERIMENTALLY ON VARIOUS SCALES
                if t==1:
                    erates[t-1]=eratesi*10**(5*np.log10((OCR[t]-1)/(OCRtarget-1)))
                ee[t]=-Cr*np.log10(OCRtarget)*UavgAnal
                e[t]=e[t-1]+(eratec[t-1]-erates[t-1])*dt+(ee[t]-ee[t-1])
                if creep_coupled==True:
                    eratec[t]=eratec[t-1]-eratec[t-1]*(eratec[t-1]-erates[t-1])*1/(Calphahatc_real[t-1]*0.434)*dt 
                else:
                    eratec[t]=eratec[t-1]-eratec[t-1]*eratec[t-1]*1/(Calphahatc_real[t-1]*0.434)*dt 
                erates[t]=erates[t-1]-erates[t-1]*erates[t-1]*1/(Calphahats*0.434)*dt 
                erate[t]=(e[t]-e[t-1])/dt#
                sigp_ref[t]=sigp_ref[t-1]+(OCR[t]-OCR[t-1])*sigmaref
                sigp_ref[t]=sigp_ref[t]+sigp_ref[t]*(eratec[t]-erates[t])/(Cr*CcCrRatio-Cr)/0.434*dt

                if creep_coupled==True:
                    Calphahatc_real[t]=Calphahatc*ratio_Calphac_f_rate(eratec[t]/erateref,beta2,beta3)/ratio_Calphac_f_rate(erateci/erateref,beta2,beta3)#
                else:
                    Calphahatc_real[t]=Calphahatc
                dt=np.clip(1/np.abs(eratec[t]-erates[t])*dtfactor,1e-100,time[t]/2)
            else:
                t = 2
                failsafe += failsafe
                dtfactor = dtfactorinit / 10
                dtmult = 1+(dtmult-1) / 2
                dtfactorinit = dtfactorinit/10
                dt=1e-100
        if failsafe > 9:
            print(Calphahatc,Calphahats,erateci,eratesi,Cr)
        return {'time':time[:t],'sigma':0,'e':e[:t],'erate':erate[:t],'erate_c':eratec[:t],'erate_s':erates[:t],'Calphahatc':Calphahatc_real[:t],'OCR':OCR[:t]}

def ratio_Calphacinit(OCR,beta2):
    return 2/((OCR)**beta2+1)
def ratio_Calphasinit(OCR,m=0.99,b=-1.44):
    return np.clip(np.clip(OCR-1,0.001,np.infty)**m*(10**b),1e-100,1e100)
def ratio_Calphac_f_rate(eratec_ratio,beta2,beta3):
    return 2/((eratec_ratio)**(-beta2/beta3)+1)
def Consol_Terzaghi_Uavg_vertical(Cv,H,t=-1,dimt=1000,dt=1,time=-1):
    '''
    Parameters
    ----------
    Cv : float
        coefficient of vertical consolidation in m/s**2
    '''
    if dimt==-1:
        time=time
    else:
        time=np.arange(0,dimt,dt)
    if t>0:
        Tv=t*Cv/((H)**2)
    else:
        Tv=time*Cv/((H)**2)
    Tv=np.clip(Tv,0,10)
    UavgAnalV=np.sqrt(4*Tv/np.pi)/((1+(4*Tv/np.pi)**2.8)**0.179)
    return time,UavgAnalV

def C_S_relaxation(Calphahatc,Calphahats,erateci,eratesi,Cc,Cr,sigma0,dimt=100000,dtfactor=1e-7,targettime=1e10,nonlinear_Calphahat=False,swelling_isotache=True,OCR = 1,beta2=4,beta3=20.5):
    time=np.zeros((dimt))
    eratec=np.zeros((dimt)); erates=np.zeros((dimt)); erate_e=np.zeros((dimt))
    e=np.zeros((dimt)); erate=np.zeros((dimt))
    sigma=np.zeros((dimt)); Calphahatc_real=np.zeros((dimt)); Calphahats_real=np.zeros((dimt))
    Calphahatc_real[0]=Calphahatc
    Calphahats_real[0]=Calphahats
    sigp=np.zeros((dimt))
    sigp[0]=sigma0*OCR
    eratec[0]=erateci
    erates[0]=eratesi
    erate[0]=0
    dt=1e-20
    sigma[0]=sigma0
    erateref=erateci/beta3_fit(beta3,OCR)
    t=0
    i=0
    failsafe=0
    dtmult = 1.01
    dtfactorinit=dtfactor
    while (time[t]<targettime) and (t<dimt-1) and (dt>0):
        if  (dt>0)&(eratec[t]>-0.01)&(erates[t]>-0.01)&(np.abs(eratec[t]*dt)<1e-2)&(np.abs(erates[t]*dt)<1e-2):
            dtfactor=dtfactor * dtmult
            t+=1
            time[t]=time[t-1]+dt   
            eratec[t]=eratec[t-1]+eratec[t-1]*erate_e[t-1]*Cc/Cr/(Calphahatc_real[t-1]*0.434)*dt #**2
            if swelling_isotache:
                erates[t]=erates[t-1]-erates[t-1]*erates[t-1]*Cc/Cr/(Calphahats_real[t-1]*0.434)*dt
            else:
                erates[t]=erates[t-1]-erates[t-1]*erates[t-1]/(Calphahats_real[t-1]*0.434)*dt

            erate_e[t]=-(eratec[t]-erates[t])
            erate[t]=erate_e[t]+eratec[t]-erates[t]
            sigma[t]=sigma[t-1]+erate_e[t]*sigma[t-1]/Cr/0.434*dt
            dt=np.clip(1/np.max([np.abs(eratec[t]),np.abs(erates[t])])*dtfactor,0,time[t]/2)
            if nonlinear_Calphahat==True:
                sigp[t]=sigp[0]/(10**((Cr/(Cc-Cr))*np.log10(sigma[t]/sigma0)))
                Calphahatc_real[t]=Calphahatc*ratio_Calphac_f_rate(eratec[t]/erateref,beta2,beta3)/ratio_Calphac_f_rate(erateci/erateref,beta2,beta3)#
                Calphahats_real[t]=ratio_Calphasinit(sigp[t]/sigma[t])*(Calphahats/ratio_Calphasinit(OCR))
            else:
                Calphahatc_real[t]=Calphahatc
                Calphahats_real[t]=Calphahats
        else:
            t = 2
            failsafe += failsafe
            dtfactor = dtfactorinit / 10
            dtmult = 1+(dtmult-1) / 2
            dtfactorinit = dtfactorinit/10
            dt=1e-100
    if failsafe > 9:
        print(Calphahatc,Calphahats,erateci,eratesi,Cr)
    return {'time':time[:t],'e':e[:t],'erate':erate[:t],'erate_c':eratec[:t],'erate_s':erates[:t],'Calphahatc':Calphahatc_real[:t],'sigma':sigma[:t],'Calphahats':Calphahats_real[:t]}

def mitsr_strain_rate(OCR,beta2,Cc,Cr,Calpha):
    return (np.sqrt(1/OCR)*np.exp((1-(OCR)**beta2)/(2*beta2)))**((Cc-Cr)/Calpha)

def beta2_fit(beta2,OCR):
    return 2/((OCR)**beta2+1)
def beta2_min_fit(beta2,minimum,OCR):
    return np.clip(2/((OCR)**beta2+1),minimum,1)
def beta3_fit(beta3,OCR):
    return (1/(OCR))**beta3
def power_law(OCR,m,b):
    return np.clip((OCR-1)**m*(10**b),1e-100,1e100)
def beta_nash(OCR,betamin,b0,a):
    b=np.clip(1/OCR,b0,1)
    mu = (np.pi/2)*((b-b0)/(1-b0))
    return np.nan_to_num(betamin+(1-betamin)*(np.sin(mu))**a)
def ratio_Calphacinit(OCR,beta2):
    return 2/((OCR)**beta2+1)
def ratio_Calphasinit(OCR,m=0.99,b=-1.44):
    return np.clip(np.clip(OCR-1,0.001,np.infty)**m*(10**b),1e-100,1e100)

def C_S_CRS(Cc,Cr,sigmap,eratec0,CalphaNC,erateref=2e-4,Cc_reduction=2.3,Cc_f_OCR=False,erates0=-1,erate = 1,sigma0=265,sigmaf=160,b2=-5.5,m2=2.5,beta2=3,beta3=18,estart=1,m1=0.88,b1=-1.4,dtfactor=1):
    Cc0 = Cc
    dimt=100000
    time=np.zeros((dimt))
    sigp_ref=np.zeros((dimt))
    eratec=np.zeros((dimt)); erates=np.zeros((dimt)); erate_e=np.zeros((dimt))
    e=np.zeros((dimt)); sigma=np.zeros((dimt))
    OCRrate=np.zeros((dimt))
    e[0]=estart
    OCR=np.zeros((dimt))
    OCRref=np.zeros((dimt))
    sigp_ref[0]=sigmap
    OCR[:]=np.clip(sigmap/sigma0,1.01,np.infty)
    OCRref[:]=np.clip(sigp_ref/sigma0,1.01,np.infty)
    dt=np.abs(1/erate/1000000)*dtfactor
    sigma[0]=sigma0
    eratec[0]=eratec0
    if erates0==-1:
        erates[0]=10**(b2)*((OCR[0]-1)**m2)
    else:
        erates[0]=erates0
    Calphahatc_real=np.zeros((dimt))
    Calphahats_real=np.zeros((dimt))
    Calphahatc_real[0]=ratio_Calphacinit(sigp_ref[0]/sigma[0],beta2)*CalphaNC
    Calphahats_real[0]=ratio_Calphasinit(sigp_ref[0]/sigma[0],m=m1,b=b1)*CalphaNC
    t=0
    stop=True
    while stop & (t<dimt-2):
        if t>1:
            if (np.abs(eratec[t]/eratec[t-1])>1.1)|(np.abs(erates[t]/erates[t-1])>1.1):
                dt=dt/1.1
            else:
                dt=dt*1.01
        t+=1
        time[t]=time[t-1]+dt
        OCR[t]=np.clip(sigmap/sigma[t-1],1.01,np.infty)
        OCRref[t]=np.clip(sigp_ref[t-1]/sigma[t-1],1.01,np.infty)
        OCRrate[t] = (OCR[t]-OCR[t-1])/dt
        eratec[t]=eratec[t-1]-eratec[t-1]*(eratec[t-1]-erates[t-1])*1/(Calphahatc_real[t-1]*0.434)*dt+eratec[t-1]*erate_e[t-1]*(Cc-Cr)/Cr*1/(Calphahatc_real[t-1]*0.434)*dt #**2
        erates[t]=erates[t-1]+m2*OCRrate[t]/(OCR[t]-1)*erates[t-1]*dt-erates[t-1]*erates[t-1]/(Calphahats_real[t-1]*0.434)*dt
        erate_e[t] = -erate-eratec[t]+erates[t]
        sigma[t]=sigma[t-1]+(erate_e[t]/Cr)*sigma[t-1]/0.434*dt
        e[t]=e[t-1]+erate*dt
        sigmap = np.max([sigma[t],sigmap])
        sigp_ref[t]=sigp_ref[t-1]+sigp_ref[t-1]*(eratec[t]-erates[t])/(Cc-Cr)/0.434*dt
        
        Calphahatc_real[t]=CalphaNC*np.clip(ratio_Calphac_f_rate(eratec[t]/erateref,beta2,beta3),0.1,100)
        #if sigmaf>sigma0:
        #    Calphahatc_real[t]=CalphaNC*np.clip(ratio_Calphac_f_rate(eratec[t]/erateref,beta2,beta3),0.1,100)
        #else:
        #    Calphahatc_real[t]=np.clip(FDM.ratio_Calphacinit(sigp_ref[t]/sigma[t],beta2),0.1,1)*CalphaNC   
        Calphahats_real[t]=np.clip(ratio_Calphasinit(OCR[t],m=m1,b=b1),0.05,100)*CalphaNC
        
        if sigma0>sigmaf:
            stop = (sigma[t]>sigmaf)
        else:
            stop = sigma[t]<sigmaf
        if Cc_f_OCR:
            Cc=Cc0-(Cc0-Cc0/Cc_reduction)/(1+np.exp(-10*(OCR[t]-1.2)))
    return {'time':time[:t],'sigma':sigma[:t],'e':e[:t],'erate_c':eratec[:t],'erate_e':erate_e[:t],'erate_s':erates[:t],'Calphahatc':Calphahatc_real[:t],'Calphahats':Calphahats_real[:t]}





import numpy as np
import matplotlib.pyplot as plt
from pynverse import inversefunc
import tqdm
import logging

from evspy.multi_stage_model.consolidation import Consol_Terzaghi_Uavg_vertical
from evspy.multi_stage_model.viscous_swelling import swelling_law
from evspy.multi_stage_model.distorted_isotaches import base_isotache_model, create_isotache, make_contour

class distorted_isotache_model(base_isotache_model):
    '''Distorted isotache model with decoupled model 

    Parameters
    ----------
    base_isotache_model : class
        Base class for isotache models
    '''
    def initialize_decoupled(self):
        self.time_step_start={}
        self.time_step_end={}
        self.time=np.array([0])
        self.sigma=np.array([self.sigma0])
        self.e[:]=self.e_init
        self.initialize_reference_curve()
        self.sigpref[0] = self.get_sigpref(sigma=self.sigma[0:1],e=self.e[0:1],sigpref=[0],initiate=True)
        if self.sigpref[0]<self.sigma0:
            self.e[0] = self.eNC[self.sigrangeNC>self.sigma0][1]
            logging.warning(f'Initial effective stress is too high for given initial void ratio; reducing to einit = {self.e[0]}')
            self.sigpref[0] = self.get_sigpref(self.sigma[0:1],self.e[0:1],sigpref=[0],initate=True)
        self.OCR_real[0]=self.sigpref[0]/self.sigma[0]
        self.OCR[0]=self.sigpref[0]/self.sigma[0]
        #print(sigma[0],e[0],beta2,beta3,Cc,Cr,CalphaNC,sigpref[0],e0,isotache,power_law,ref_func)
        self.eratec[0]=self.calc_pos_isotache(self.sigma[0],self.e[0],self.sigpref[0])
        self.erates[0]=self.erateinits
        self.erate[0]=self.eratec[0]-self.erates[0]
        self.dt=1e-20#np.clip(1/np.abs(erate[0])*dtfactor,1e-5,1e-3)
        self.Calphahatc_real=np.zeros((self.dimt))
        self.Calphahats_real=np.zeros((self.dimt))
        self.OCRrate=np.zeros((self.dimt))
        self.Calphahats_real[0]=self.CalphaCc*self.Cc**10**self.b1*(self.OCR[0]-1)**self.m1
        
        #for t in tqdm.tqdm(range(1,dimt)):

    def initiate_new_load_step(self,load_step,t,running,reset_minimum_swelling_rate):
        next_load_step=False
        if self.load.type[load_step]=='IL':
            if (t>=len(self.time)-1): #(time[t]>=load.end_time[load_step])|
                #print(load_step)
                next_load_step=True
        elif self.load.type[load_step]=='CRS':
            if self.load.rate[load_step]>0:
                if self.sigma[t-1]>=self.load.load[load_step]:
                    next_load_step=True
            elif self.load.rate[load_step]<0:
                if self.sigma[t-1]<=self.load.load[load_step]:
                    next_load_step=True
            else:
                if (self.time[t]-self.time_step_end[self.load.type[load_step-1]+str(load_step-1)]>=self.load.duration[load_step]):
                    next_load_step=True
        if next_load_step:
            reset_minimum_swelling_rate = 0
            if len(self.load.type)>load_step+1:
                #print('next step '+str(load_step))
                self.time_step_end[self.load.type[load_step]+str(load_step)]=self.time[t]
                load_step+=1
                if self.load.type[load_step]=='IL':
                    time_u,U=Consol_Terzaghi_Uavg_vertical(self.load.cv[load_step]/3600/24/365,self.H,targettime=self.load.duration[load_step],dimt=1000,constant_dt_time=1e6,dtmax=1e5)
                    U=np.append(U[0],np.clip(U[1:]-U[:-1],0,1).cumsum())
                    self.time=np.append(self.time,self.time[-1]+0.01+time_u)
                    self.sigma=np.append(self.sigma,self.sigma[-1]+1/U[-1]*(self.load.load[load_step]-self.sigma[-1])*np.clip(U,0,1))
                elif self.load.type[load_step]=='CRS':
                    self.erate_crs = self.load.rate[load_step]
                self.time_step_start[self.load.type[load_step]+str(load_step)]=self.time[t]
            else:
                running=False
        return next_load_step, running, load_step, reset_minimum_swelling_rate
         
    def run_iterations(self,timefactor=1.2):
        t = 0
        #start=False
        i = 0
        failsafe=0
        dtmult = 1.01
        targettime=1e7
        dtfactor=1
        dtfactorinit=dtfactor
        OCR_ref=self.OCR_real[0]
        self.sigp_update[0]=self.sigpref[0]
        #while (t<dimt-1) and (dt>0): #(time[t]<targettime) and 
        printing=False
        load_step = 0
        next_load_step=True
        running=True
        reset_minimum_swelling_rate = 0
        while (running)&(t<self.dimt-1):
            next_load_step, running, load_step, reset_minimum_swelling_rate=self.initiate_new_load_step(load_step,t,running, reset_minimum_swelling_rate)
            if running:
                t+=1
            if self.load.type[load_step]=='CRS':
                if next_load_step:
                    dt = 1e-10
                else:
                    dt = np.clip(np.abs(1/self.erate_e[t-1])/15000,0.001,np.min([dt*timefactor,5e4]))
                self.time=np.append(self.time,[self.time[t-1]+dt])
                self.erate_e[t] = self.erate_crs+self.eratec[t-1]+self.erates[t-1]
                self.sigma=np.append(self.sigma,[self.sigma[t-1]+((10**(self.erate_e[t]*dt/self.Cr)-1)/dt)*self.sigma[t-1]*dt])
                self.ee[t]=self.ee[t-1]-self.erate_e[t]*dt 
            elif self.load.type[load_step]=='IL':
                dt=self.time[t]-self.time[t-1]
                self.erate_e[t]= self.Cr*0.434*(self.sigma[t]-self.sigma[t-1])/dt/np.average([self.sigma[t],self.sigma[t-1]])
                self.ee[t]=self.ee[t-1]-self.erate_e[t]*dt 
            
            self.OCR_real[t]=self.sigpref[t-1]/self.sigma[t]
            self.OCR[t]=np.clip(np.max(np.append(np.array([self.sigpref[0]]),self.sigma[:t]))/self.sigma[t],1.005,np.infty)
            self.OCRrate[t] = (self.OCR[t]-self.OCR[t-1])/dt
            
            sigmarate=(self.sigma[t-1]-self.sigma[t-2])/dt
            self.erates[t],self.Calphahats_real[t-1],reset_minimum_swelling_rate=swelling_law(self.erates[t-1],self.OCR[t],self.OCRrate[t],
                                                                        self.Calphahats_real[t-1],dt,self.CalphaCc*self.Cc,
                                                                        self.m1,self.b1,self.m2,self.b2,sigmarate,-self.erate_e[t],reset_minimum_swelling_rate)
            
            # Distortion occurs only during swelling and not during relaxation
            if ((self.sigma[t]<self.sigma[t-1]/1.000001)
                    &(self.sigma[t-1]<self.sigma[t-2]/1.000001)
                    &(np.abs(self.e[t]-self.e[t-1])>0)
                    &(np.abs(self.e[t-2]-self.e[t-1])>0)):
                if (self.load.type[load_step]=='IL')&(np.round(self.load.load[load_step-1],1)==np.round(self.load.load[load_step],1)):
                    self.sigp_update[t]=self.sigp_update[t-1]
                else:
                    self.sigp_update[t]=np.max(self.sigpref[self.sigpref>0])
            else:
                self.sigp_update[t]=self.sigp_update[t-1]

            self.e[t]=self.e[t-1]+(self.eratec[t-1]+self.erates[t-1])*dt+(self.ee[t]-self.ee[t-1])
            self.sigpref[t] = self.get_sigpref(self.sigma[t-1:t+1],self.e[t-1:t+1],self.sigpref[t-1:t])
            self.eratec[t]=self.calc_pos_isotache(self.sigma[t],self.e[t],self.sigp_update[t],steps=20)
            self.eratec[t]=-self.eratec[t]
            
            if np.isnan(self.ee[t]):
                print(dt)
                print(self.Cr*0.434*(self.sigma[t]-self.sigma[t-1])/dt/np.average([self.sigma[t],self.sigma[t-1]]))
                print(str(self.sigma[t-1])+' '+str(self.sigma[t]))
                print('failure')
            if self.eratec[t]>1:
                print((self.sigma[t],self.e[t],self.beta2, self.beta3,self.Cc,self.Cr,self.CalphaNC,self.sigp_update[t],self.e0))
            self.erate[t]=(self.e[t]-self.e[t-1])/dt#
            
            self.Calphahats_real[t]=self.CalphaCc*self.Cc*10**self.b1*(np.clip(self.OCR[t]-1,0.005,np.infty))**self.m1

        self.time = self.time[:t]
        self.sigma = self.sigma[:t]
        self.e = self.e[:t]
        self.erate = self.erate[:t]
        self.Calphahats_real = self.Calphahats_real[:t]
        self.sigpref = self.sigpref[:t]
        self.eratec = self.eratec[:t]
    
    def plotting_stresspath(self,sigp):
        dforg,eUC_org,eUC0_org,eNC_org,sigrange_org,sigrangeNC=create_isotache(erateref = self.erateref, beta2 = self.beta2, sigp=sigp,
                                                                                beta3 = self.beta3,Cc = self.Cc,Cr = self.Cr,CalphaNC = self.CalphaNC,
                                                                                e0=self.e0,isotache=self.isotache, power_law=self.use_power_law, 
                                                                                ref_func=self.ref_func, dsig=self.dsig, beta_nash=self.use_beta_nash,param_nash=self.param_nash)
        dforg=dforg.reset_index()
        cs=make_contour(dforg,eUC_org,eUC0_org,eNC_org,sigrange_org,sigrangeNC,num=20,figsize=(8,5),vmax=-3,colorbar=True)
        plt.plot(self.sigma[0],self.e[0],'o')
        plt.plot(self.sigma,self.e,'k-',lw=4)
        plt.ylim(np.min(self.e),np.max(self.e))
        plt.xlim(1,1000)
        #df,eUC,eUC0,eNC,sigrange,sigrangeNC=create_isotache(beta2 = beta2, beta3 = beta3,Cc = Cc,Cr = Cr,CalphaNC = CalphaNC,sigp=sigpref[t-1],e0=e0,isotache=isotache,power_law=power_law,ref_func=ref_func,dsig=0.05,beta_nash=beta_nash)
        #cs=make_contour(df,eUC,eUC0,eNC,sigrange,sigrangeNC,num=20,figsize=(8,5),vmax=-3)
        #plt.plot(sigma[0],e[0],'o')
        #plt.plot(sigma[:t],e[:t],lw=4)
        #plt.ylim(1,2.2)
        #plt.xlim(1,500)
    #print(load_step)
    #return time[:t],sigma[:t],e[:t],erate[:t],erate_e[:t],eratec[:t],erates[:t],Calphahats_real[:t],ee[:t],sigpref[:t],time_step,OCR[:t],OCR_real[:t],sigp_update[:t]

    def plotting_timeseries(self,loadstep : str, set_zero=False,**kws):
        time = self.time[(self.time>self.time_step_start[loadstep])&(self.time<self.time_step_end[loadstep])]-self.time[(self.time>=self.time_step_start[loadstep])][0]
        e = self.e[(self.time>self.time_step_start[loadstep])&(self.time<self.time_step_end[loadstep])]
        sigma = self.sigma[(self.time>self.time_step_start[loadstep])&(self.time<self.time_step_end[loadstep])]
        if set_zero:
                e0 = e[0]
                sigma0 = sigma[0]
        else:
            e0 = 0
            sigma0 = 0
        if 'IL' in loadstep:
            plt.semilogx(time,e-e0,**kws)
            plt.xlabel('Time (s)')
            plt.ylabel('Void ratio $e$')
        else:
            plt.semilogx(time,sigma-sigma0,**kws)
            plt.xlabel('Time (s)')
            plt.ylabel('Stress (kPa)')

if __name__ == '__main__':
    from evspy.loadsteps import LoadSteps
    load=LoadSteps()
    load.add_load_step(1e5,5,1e-8,'CRS',cv=7)
    #load.add_load_step(1e5,15,1e-6,'CRS',cv=7)
    #load.add_load_step(1e5,30,1e-8,'CRS',cv=7)
    #load.add_load_step(1e5,70,1e-6,'CRS',cv=7)
    #load.add_load_step(1e5,100,1e-8,'CRS',cv=7)
    #load.add_load_step(1e5,200,1e-8,'IL',cv=7)
    #load.add_load_step(1e8,200,1e-8,'IL',cv=7)
    #load.add_load_step(1e8,20,-1e-6,'CRS',cv=7)
    #load.add_load_step(5e6,20,-1e-8,'IL',cv=7)
    #load.add_load_step(1e7,20,0,'CRS',cv=7)
    model = distorted_isotache_model(load,sigma0=1,H=0.1,dimt=100)
    # Add a check to see if initial condition is reasonable (for instance not above reference line)
    model.initialize_decoupled(1,1,1e-10)
    model.run_iterations()


from setuptools import setup, find_packages

# load the README file and use it as the long_description for PyPI
def readme():
    with open('README.md', 'r') as f:
        return f.read()

setup(name='evspy',
      version='0.2.0',
      description='Modelling of creep and swelling during unloading',
      long_description=readme(),
      url='http://github.com/thomasvergote/evspy',
      author='Thomas Vergote',
      author_email='thomas@inferensics.be',
      license='GPL v3',
      packages=find_packages(),
      keywords=['engineering', 'geotechnical'],
      classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
      ],
      zip_safe=False)



import datetime
import pandas as pd
import numpy as np
from evspy.single_stage_model.model_run import decoupled_cs_model

def test_decoupled_CS_model():
    Calphahatc = 0.0015
    Calphahats = 0.0005
    erateci = 1e-8
    eratesi = 1e-6
    Cr = 0.05
    Cc = Cr*5
    CalphaNC = Cc * 0.04
    Cv = 2/365/3600
    OCR_initial = 1.0
    OCR_final = 1.5
    model = decoupled_cs_model(Calphahatc, Calphahats,  erateci, eratesi, Cc,Cr,CalphaNC, OCR_initial,OCR_final,Cv=Cv,nonlinear_Calphahat=True,hypA=False,beta2=4,beta3=20)
    
    assert np.round(model.OCR[-1],2) == OCR_final


from evspy.utils.helper_functions import semilogintersection, loglogintersection, line_intersection
import numpy as np

def test_line_intersection():
    line1 = ((0.5,1),(1.5,0.5))
    line2 = ((1,0),(2,2))
    x,y = line_intersection(line1,line2)
    assert x == 1.3
    assert y == 0.6

def test_semilogintersection():
    e0 = 2
    e1 = 1
    sigma1 = 1
    Cc = 1
    Cr = 0
    x,y = semilogintersection(e0, e1, sigma1, Cc, Cr)
    assert x == 10
    assert y == 1

def test_loglogintersection():
    e0 = 2
    e1 = 1
    sigma1 = 1
    rhoc = 1
    Cr = 0
    x = loglogintersection(e0, e1, sigma1, rhoc, Cr)
    assert np.round(x[0],1) == 2


import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from pynverse import inversefunc
import lmfit

from scipy.interpolate import griddata
from scipy.special import lambertw
from scipy.optimize import minimize, curve_fit

# Intersections

def semilogintersection(e0,e1,sigma1,Cc,Cr):
    '''Intersection between two compression curves using semilogarithmic relations

    Parameters
    ----------
    e0 : float
        reference void ratio (y-axis) of compression curve
    e1 : float
        void ratio (y-axis) at current stress point
    sigma1 : float
        stress (x-axis) at current stress point
    Cc : float
        compression index
    Cr : float
        recompression index

    Returns
    -------
    tuple of floats
        coordinates at intersection
    '''
    m1=-Cc
    m2=-Cr
    b1=e0
    b2=e1+Cr*np.log10(sigma1)
    xi = (b1-b2) / (m2-m1)
    yi = m1 * xi + b1
    xi = 10**xi
    return xi,yi
semilogintersection=np.vectorize(semilogintersection)

def loglogintersection(e0,e1,sigma1,rhoc,Cr):
    '''Intersection between two compression curves using log-log relations

    Parameters
    ----------
    e0 : float
        reference void ratio (y-axis) of compression curve
    e1 : float
        void ratio (y-axis) at current stress point
    sigma1 : float
        stress (x-axis) at current stress point
    rhoc : float
        log-log compression index
    Cr : float
        recompression index
    
    Returns
    -------
    float
        x-coordinate at intersection
    '''
    def f(sigma,rhoc,Cr,e0,e1,sigma1):
        return np.abs(np.exp(np.log(e0)-rhoc*np.log(sigma))-e1+Cr*np.log10(sigma/sigma1))
    return minimize(f,x0=sigma1,args=(rhoc,Cr,e0,e1,sigma1))['x']

def line_intersection(line1, line2):
    '''Finds intersection between two lines given end points of the lines

    Parameters
    ----------
    line1 : tuple
        ((x0,y0),(x1,y1)) coordinates of end points of first line
    line2 : tuple
        (x0,y0),(x1,y1)) coordinates of end points of second line

    Returns
    -------
    tuple
        intersection point
    '''
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    #if div == 0:
       #raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y


def get_intersection(sigrangeUC,eUC,sigrangeNC,eNC,Cr):
    y=(eUC-Cr*np.log10(sigrangeNC[:,None]/sigrangeUC[None,:]))
    idx=pd.DataFrame(np.argwhere(np.diff(np.sign(y.transpose() - eNC)))).drop_duplicates(subset=[0])
    idx.index=idx[0]
    df=pd.DataFrame()
    df['target']=np.linspace(0,len(eUC)-1,len(eUC))
    idx=pd.concat([df,idx],axis=1).interpolate()
    idx=idx.drop('target',axis=1).values
    idx=idx.astype('int')
    return sigrangeNC[idx[:,1]], eNC[idx[:,1]]

def make_grid(df,sigrange,de=0.0002):
    sigi=sigrange
    ei=10**np.arange(np.log10(0.5),np.log10(3),de)
    erateii=griddata((df['sigma'],df['e']),np.log10(df['erate']),xi=(sigi[None,:],ei[:,None]))
    return sigi,ei,erateii

def linfit(x,a,b):
    return a*x+b

def semilogfit(x,a,b):
    return a*np.log10(x)+b

def loglogfit(x,a,b):
    return 10**(a*np.log10(x)+b)

def fit_Cc_Cr(xrange,yrange,function=semilogfit,label=r'$C_c$ = '):
    xrange=xrange[(xrange>-1000)&(yrange>-1000)]
    yrange=yrange[(xrange>-1000)&(yrange>-1000)]
    popt,pcov = curve_fit(function,xrange,yrange)
    plt.plot(10**np.arange(-1,3,0.1),function(10**np.arange(-1,3,0.1),popt[0],popt[1]),'k--',label=label+str(np.round(-popt[0],2)))
    return -popt[0],popt[1]

def get_sigref(sigma,e,Cc,Cr,e0):
    X,Y=line_intersection(((np.log10(sigma),e),(5,e-Cr*np.log10(10**5/sigma))),((0,e0),(5,e0-Cc*5)))
    return 10**X,Y